c-----------------------------------------------------------------------
      program prepaq
c-----------------------------------------------------------------------
c
c --- PREPAQ    Version: 1.18           Level: 110225              MAIN
c
c               D. Strimaitis
c
c     Copyright (c) 2006-2011 by Exponent, Inc.
c
c --- PURPOSE: Convert a CSV file of hourly AQ data to the data
c              record section of the TRAJLOC.DAT file for TRAJ2D.FOR.
c
c --- UPDATES:
c --- V1.17 (100128) to V1.18 (110225) DGS
c       (1) Updated CALUTILS.FOR to version v2.58 (110225)
c           - Add control file variable type 5 (character array) and
c             retain commas in string returned so that array of
c             variable values can be parsed
c           Modified: READIN, ALTONU, SETVAR
c --- V1.16 (100108) to V1.17 (100128) CEC
c       (1) Make a correction in checking empty VALID flag to allow some
c           valid column to be partially empty.
c           Modified: READPD 
c       (2) Add a check when any possible concentration is empty
c           or have already missing parameters such as -6999. or 9999. or -9999.
c           Modified: RGET320
c --- V1.15 (091016) to V1.16 (100108) CEC
c       (1) A character in non standard ASCII is used for "space".
c           in AMMNET files format. A Fix is to accept as VALID 
c           any FLAGS characters which are not unvalid.
c           Modified: VGET320
c
c --- v1.14 (091005) to v1.15 (091016) DGS
c         - Identify cases where requested data are not found
c         - Add more information to listfile error messages
c         - Add completion text string to list file that can be used by
c           GUI to check for proper termination:
c           'RUN COMPLETED SUCCESSFULLY'
c           Modified:  COMP, READPD, FIN
c         - Use RD_DATIM subroutine to interpret several date-time
c           formats from spreadsheet CSV field
c           New:       RD_DATIM
c           Modified:  READPD
c                      
c
c --- v1.13 (090310) to v1.14 (091005) DGS
c         - Trap case of void input data in CSV file, and treat as
c           missing
c         - Add MVALID switch for ignoring validation data in CSV file
c           Modified:  CONTROL.PAQ
c                      BLOCKDATA, READCF, READHD, READPD
c
c --- v1.12 (081205) to v1.13 (090310) DGS
c         - Add MPROC option to create output records for all periods
c           for which concentrations exceed a threshold (or retain
c           daily peak method)
c         - CALUTILS updated from v2.56 to v2.57 to allow control file
c           lines as long as 200 characters and to activate CPU time
c         - Correct daily time-window logic that did not work when
c           the end of the window is at midnight
c         - Recognize 9999 as a missing concentration in AMMNET files
c           if explicit validity codes are not available
c           Modified:  CONTROL.PAQ, PARAMS.CAL, CALUTILS.FOR
c                      BLOCKDATA, READCF, COMP, READHD, READPD
c
c --- v1.11 (081117) to v1.12 (081205) DGS
c         - Change default output file extension to TAQ
c         - Add concentration threshold
c           Modified:  CONTROL.PAQ
c                      BLOCKDATA, READCF, READPD
c         - Change processor name to PREPAQ (with .PAQ include files)
c
c --- v1.1 (081017) to v1.11 (081117) DGS
c         - Interpret data validity flags in AMMNET-format file
c           New:       VGET320
c           Modified:  AQINP.PAQ, CONTROL.PAQ
c                      BLOCKDATA, READHD, READPD
c
c --- v1.0 (061231) to v1.1 (081017) DGS
c         - Place program inputs into standard control file
c         - Remove HDR input file by placing content in control file
c
c --- PREPAQ calls:     SETUP, COMP, FIN
c-----------------------------------------------------------------------
c --- Include parameters
      include 'params.paq'
c --- Include common blocks
      include 'qa.paq'

c --- Set version and level number of program
      ver='1.18'
      level='110225'

c --- SETUP PHASE -- read control file information
      call SETUP

c --- COMPUTATIONAL PHASE -- loop over processing periods
      call COMP

c --- TERMINATION PHASE -- program termination functions
      call FIN(2)

      stop
      end

c----------------------------------------------------------------------
      BLOCK DATA
c----------------------------------------------------------------------
c
c --- PREPAQ     Version: 1.16        Level: 091005          BLOCK DATA
c                D. Strimaitis
c
c --- Include parameter statements
      include 'params.paq'
c
c --- Include common blocks
      include 'filnam.paq'
      include 'control.paq'
      include 'qa.paq'

c --- FILNAM common block
      data runinp/'prepaq.inp'/,runlst/'prepaq.lst'/,
     1 aqdat/'aqual.dat'/,taqdat/'prepaq.taq'/
c --- FILLOG common block
      data lcfiles/.false./

c --- CONTROL common block
      data itype/2/, mproc/0/, mvalid/1/
      data ihr1/1/, ihr2/24/, idelpd1/0/, idelpd2/0/
      data col_dtm/'DATETIME'/
      data col_avg/'DATAAVID'/
      data col_stn/'STNCODE'/
      data col_vld/'VALID'/
      data pmap/'UTM     '/, utmhem/'N   '/
      data datum/'WGS-84  '/, daten/'02-21-2003  '/
      data feast/0.0/, fnorth/0.0/
      data lgeo/.FALSE./
      data lutm/.FALSE./
      data llcc/.FALSE./
      data lps/.FALSE./
      data lem/.FALSE./
      data llaza/.FALSE./
      data lttm/.FALSE./
      data cminppb/1./

c --- QA common block
      data model/'PREPAQ      '/

      end

c----------------------------------------------------------------------
c --- BRING IN CALPUFF SYSTEM UTILITY SUBROUTINES
      include 'calutils.for'
c----------------------------------------------------------------------

c----------------------------------------------------------------------
      subroutine setup
c----------------------------------------------------------------------
c
c --- PREPAQ     Version: 1.18       Level: 081017                SETUP
c                D. Strimaitis
c
c --- PURPOSE:   perform all initialization and one-time setup
c                operations
c
c     Common block /FILNAM/
c        RUNINP
c
c --- Parameters used:
c        ICTL, ILST, IOMESG
c
c --- SETUP called by: MAIN
c --- SETUP calls:     DATETM, COMLINE, READCF, RDHEAD
c                      
c----------------------------------------------------------------------
c
c --- Include parameters
      include 'params.paq'

c --- Include common blocks
      include 'filnam.paq'
      include 'qa.paq'

c --- Get date and time from system
      call DATETM(rdate,rtime,rcpu)

c --- Get the name of the control file from the command line
      call COMLINE(runinp)

c --- Open the control file
      open(ictl,file=runinp,status='old')

c --- Report progress
      write(iomesg,*)'PREPAQ Processor Running'
      write(iomesg,*)'SETUP PHASE'

c --- Check that the version and level number in the parameter
c --- file matches those in the code itself
      if(ver.ne.mver.or.level.ne.mlevel)then
         write(iomesg,10) ver,level,mver,mlevel
10       format(/1x,'ERROR in SUBR. SETUP -- The PREPAQ version ',
     1   'and level numbers do not match those in the parameter file'/
     2   5x,'    Model Code - Version: ',a12,' Level: ',a12/
     3   5x,'Parameter File - Version: ',a12,' Level: ',a12)
         stop
      endif

c --- Read control file (open other files)
      call READCF

c --- Process header records of input data file
      call READHD

      return
      end

c----------------------------------------------------------------------
      subroutine readhd
c----------------------------------------------------------------------
c
c --- PREPAQ     Version: 1.18       Level: 091005               READHD
c                D. Strimaitis
c
c --- PURPOSE:   Read and process header of Type 2 AQ file
c
c --- UPDATES:
c --- v1.13 (090310) to v1.14 (091005) DGS
c         - Add option to ignore validity codes
c --- v1.11 (081117) to v1.13 (090310) DGS
c         - Recognize 9999. as missing if validity flags are not in
c           AMMNET-format file
c --- v1.1 (081017) to v1.11 (081117) DGS
c         - Add validity flags in AMMNET-format file
c
c --- INPUTS:
c
c ---    Common block /CONTROL/ variables:
c           ITYPE,CSPEC,col_dtm,col_avg,col_stn,col_vld,mvalid
c
c --- Parameters used:
c        ICSV, ILST
c
c --- OUTPUT:
c
c ---    Common block /AQINPUT/ variables:
c           ICOLDATM, ICOLSTN, ICOLAVG, ICOLCONC, ICOLVLD, NVARS
c
c --- READHD called by: SETUP
c --- READHD calls:     OMIT320, COLUMN320
c----------------------------------------------------------------------
c
c --- Include parameters
      include 'params.paq'

c --- Include common blocks
      include 'control.paq'
      include 'aqinput.paq'

c --- Local variables
      integer icommas(0:99)
      logical lerror,lwarn
      character*1 cm,dq
      character*320 char320,blnk320

      data cm/','/
      data dq/'"'/
      data lerror/.FALSE./,lwarn/.FALSE./

      do i=1,320
         blnk320(i:i)=' '
      enddo

c --- Check for expected file type
      if(itype.EQ.1) then
         return
      elseif(itype.NE.2) then
         write(*,*)'READHD:  Error found'
         write(*,*)'File type identified in control file = ',itype
         write(*,*)'File type expected for READHD = 2'
         stop
      endif

c --- Prepare Species name length
      kspec=LEN_TRIM(cspec)

c --- Read line
      char320=blnk320
      read(icsv,'(a320)') char320

c --- Create upper-case version of line
      klast=LEN_TRIM(char320)
      call CASE320('UPPER',char320,klast)

c --- Set column name lengths
      kdtm=LEN_TRIM(col_dtm)
      kavg=LEN_TRIM(col_avg)
      kstn=LEN_TRIM(col_stn)
      kvld=LEN_TRIM(col_vld)

c --- Check for required field names
c --- (VALID field may not appear)
      idatm=INDEX(char320,col_dtm(1:kdtm))
      if(idatm.EQ.0) lerror=.true.
      iaver=INDEX(char320,col_avg(1:kavg))
      if(iaver.EQ.0) lerror=.true.
      istcd=INDEX(char320,col_stn(1:kstn))
      if(istcd.EQ.0) lerror=.true.
      iconc=INDEX(char320,cspec(1:kspec))
      if(iconc.EQ.0) lerror=.true.
      ivld=INDEX(char320,col_vld(1:kvld))
      if(ivld.EQ.0 .AND. mvalid.NE.0) lwarn=.true.
      if(LERROR) then
         write(*,*)
         write(*,*)'READHD:  Error found'
         write(*,*)'Header record does not contain required fields'
         write(*,*)'See list file'
         write(ilst,*)
         write(ilst,*)'READHD:  Error found'
         write(ilst,*)'Header record does not contain required fields'
         write(ilst,*)'Header:  '//char320(1:klast)
         write(ilst,*)'Expected: '
         write(ilst,*) col_dtm
         write(ilst,*) col_avg
         write(ilst,*) col_stn
         write(ilst,*) cspec
         if(mvalid.NE.0) write(ilst,*) col_vld
         write(ilst,*)'idatm,iaver,istcd,iconc,ivld = ',
     &                 idatm,iaver,istcd,iconc,ivld
         stop
      elseif(LWARN) then
         write(ilst,*)
         write(ilst,*)'READHD:  WARNING'
         write(ilst,*)'Header record does not contain requested fields'
         write(ilst,*)'Header:  '//char320(1:klast)
         write(ilst,*)'Expected: '
         write(ilst,*) col_vld
         write(ilst,*)'ivld = ',ivld
      endif

c --- Drop any double quotes from line
      call OMIT320(dq,char320)

c --- Identify column ranges (NVARS=99 here because it is unknown)
      nvars=99
      klast=LEN_TRIM(char320)
      call COLUMN320(char320,klast,cm,nvars,ncommas,icommas)

c --- Step through fields to identify needed columns
      icoldatm=0
      icolavg=0
      icolstn=0
      icolconc=0
      icolvld=0
      do k=1,nvars
         call IRANGE(k,icommas,ncommas,i1,i2)
         if(char320(i1:i2).EQ.col_dtm(1:kdtm)) icoldatm=k
         if(char320(i1:i2).EQ.col_avg(1:kavg)) icolavg=k
         if(char320(i1:i2).EQ.col_stn(1:kstn)) icolstn=k
         if(char320(i1:i2).EQ.cspec(1:kspec)) icolconc=k
         if(char320(i1:i2).EQ.col_vld(1:kvld)) icolvld=k
      enddo
c --- QA

c --- Confirm columns in AQ file
      write(ilst,*)
      write(ilst,*)'Data columns in AQ file'
      write(ilst,*)'Date-Time      (icoldatm) = ',icoldatm
      write(ilst,*)'Averaging Time (icolavg)  = ',icolavg
      write(ilst,*)'Station ID     (icolstn)  = ',icolstn
      write(ilst,*)'Concentration  (icolconc) = ',icolconc
      write(ilst,*)'Validity Flag  (icolvld)  = ',icolvld
      write(ilst,*)

      if(icoldatm.EQ.0) then
         write(*,*)
         write(*,*)'READHD:  Error found'
         write(*,*)'Column not found for DATE-TIME'
         stop
      endif
      if(icolavg.EQ.0) then
         write(*,*)
         write(*,*)'READHD:  Error found'
         write(*,*)'Column not found for averaging period'
         stop
      endif
      if(icolstn.EQ.0) then
         write(*,*)
         write(*,*)'READHD:  Error found'
         write(*,*)'Column not found for Station ID'
         stop
      endif
      if(icolconc.EQ.0) then
         write(*,*)
         write(*,*)'READHD:  Error found'
         write(*,*)'Column not found for Concentration'
         stop
      endif
      if(mvalid.EQ.1) then
         if(icolvld.EQ.0) then
            write(*,*)
            write(*,*)'READHD:  Warning!'
            write(*,*)'Column not found for Validity Flags'
            write(*,*)'All values are considered valid'
            write(*,*)'unless value=9999.'
            write(*,*)
            write(ilst,*)
            write(ilst,*)'READHD:  Warning!'
            write(ilst,*)'Column not found for Validity Flags'
            write(ilst,*)'All values are considered valid'
            write(ilst,*)'unless value=9999.'
         endif
      else
         if(icolvld.GT.0) then
            write(*,*)
            write(*,*)'READHD:  Warning!'
            write(*,*)'Validity Flags are present but ignored'
            write(*,*)'All values are considered valid'
            write(*,*)'unless value=9999.'
            write(*,*)
            write(ilst,*)
            write(ilst,*)'READHD:  Warning!'
            write(ilst,*)'Validity Flags are present but ignored'
            write(ilst,*)'All values are considered valid'
            write(ilst,*)'unless value=9999.'
         endif
      endif

      return
      end

c----------------------------------------------------------------------
      subroutine readpd(ieof)
c----------------------------------------------------------------------
c
c --- PREPAQ     Version: 1.18       Level: 100128               READPD
c                D. Strimaitis
c
c --- PURPOSE:   Read and process data record of Type 1 or 2 AQ file
c                Integer concentrations in Type 1 are PPB
c                Real concentrations in Type 2 are PPM
c
c --- UPDATES:
c --- v1.15 (091016) to v1.17 (100128)
c       (1) Make a correction in checking empty VALID flag to allow some
c           valid column to be partially empty.
c --- v1.14 (091005) to v1.15 (091016) DGS
c         - Return rconc=0.0 if valid but less than threshold
c         - Use RD_DATIM subroutine to interpret several date-time
c           formats
c --- v1.13 (090310) to v1.14 (091005) DGS
c         - Trap void in CSV file as a missing
c         - Add MVALID option to ignore validity codes
c --- v1.12 (081205) to v1.13 (090310) DGS
c         - Recognize 9999. as missing if validity flags are not in
c           AMMNET-format file
c --- v1.11 (081117) to v1.12 (081205) DGS
c         - Minimun concentration provided in control file
c --- v1.1 (081017) to v1.11 (081117) DGS
c         - Add validity flags in AMMNET-format file; remove shift to
c           upper case
c
c --- INPUTS:
c
c ---    Common block /CONTROL/ variables:
c           ITYPE, CMINPPB, MVALID
c ---    Common block /AQINPUT/ variables:
c           ICOLDATM, ICOLSTN, ICOLAVG, ICOLCONC, ICOLVLD
c
c --- Parameters used:
c        ICSV, ILST
c
c --- OUTPUT:
c          IEOF - integer    - End-of-File flag
c                              0 = NO
c                              1 = Yes
c
c ---    Common block /AQINPUT/ variables:
c           iavg,istn,iyr,jul,ihr,isec,rconc
c
c --- READPD called by: COMP
c --- READPD calls:     JULDAY, RD_DATIM
c                       OMIT320, COLUMN320, IRANGE, RGET320, VGET320
c----------------------------------------------------------------------
c
c --- Include parameters
      include 'params.paq'

c --- Include common blocks
      include 'control.paq'
      include 'aqinput.paq'

c --- Local variables
      integer icommas(0:99), iconc
      integer ioffset
      logical lvalid
      logical lerror
      character*1 cm,dq
      character*16 datm
      character*320 char320,blnk320

      data cm/','/
      data dq/'"'/
      data lerror/.FALSE./

      data t1_scale_ppb/1.0/
      data t2_scale_ppb/1000.0/
      data zero/0.0/

c --- Data column offset for interpreting validity code
c --- Validity codes are placed into a string using 1 character per
c --- data column (1st character in string is for first data), but
c --- no information in file identifies the first column of data.
c --- Example file places 1st data in column 4.
      data ioffset/3/

      do i=1,320
         blnk320(i:i)=' '
      enddo

c --- Check for expected file type
      if(itype .LT.1 .OR. itype.GT.2) then
         write(*,*)'READHD:  Error found'
         write(*,*)'File type identified in control file = ',itype
         write(*,*)'File type expected for READPD = 1 or 2'
         stop
      endif

c --- Read TYPE=1 Data Record
      if(itype.EQ.1) then
         read(icsv,*,end=999) iyr,jul,ihr,iconc
         isec=0
         istn=idstn
         iavg=3600
         rconc=t1_scale_ppb*FLOAT(iconc)
      endif

c --- Read TYPE=2 Data Record
      if(itype.EQ.2) then

c ---    Read line
20       char320=blnk320
         read(icsv,'(a320)',end=999) char320

c ---    Test line for at least 1 comma to screen out empty lines
         itest=INDEX(char320,cm)
         if(itest.EQ.0) goto 20

c ---    Remove this shift since validity flags are case-sensitive
cc ---    Create upper-case version of line
c         klast=LEN_TRIM(char320)
c         call CASE320('UPPER',char320,klast)

c ---    Drop any double quotes from line
         call OMIT320(dq,char320)

c ---    Identify column ranges
         klast=LEN_TRIM(char320)
         call COLUMN320(char320,klast,cm,nvars,ncommas,icommas)

c ---    Step through fields
c --------------------------

c ---    Date-Time
         datm='mm/dd/yyyy hh:mm'
         call IRANGE(icoldatm,icommas,ncommas,i1,i2)
         datm=char320(i1:i2)
c         read(datm(7:10),'(i4)')  iyr
c         read(datm(1:2),'(i2)')   imo
c         read(datm(4:5),'(i2)')   iday
c         read(datm(12:13),'(i2)') ihr
c         read(datm(15:16),'(i2)') imm
         call RD_DATIM(i1,i2,datm,imo,iday,iyr,ihr,imm)
         isec=imm*60
c ---    Skip blank record (based on date)
         if(iyr.EQ.0 .OR. imo.EQ.0 .OR. iday.EQ.0) goto 20
         call JULDAY(ilst,iyr,imo,iday,jul)

c ---    Averaging Time
         call IRANGE(icolavg,icommas,ncommas,i1,i2)
         read(char320(i1:i2),'(i)') iavg
         iavg=iavg*60

c ---    Station Code
         call IRANGE(icolstn,icommas,ncommas,i1,i2)
         read(char320(i1:i2),'(i)') istn

c ---    Concentration
         call IRANGE(icolconc,icommas,ncommas,i1,i2)
c ---    Trap void
         if(i2.LT.i1) then
            rconc=-1.
         else
            call RGET320(char320,i1,i2,rconc)
         endif
c ---    Treat all value larger than 9999. or smaller than -9999. as missing values
         if(rconc.GT.9998. .OR. rconc.LT.-9998.) rconc=-1.

c ---    Validity of concentration
         if(icolvld.GT.0 .AND. mvalid.EQ.1) then
            call IRANGE(icolvld,icommas,ncommas,i1,i2)
c ---       Exclude void or short validity string
c ---       Remove ioffset number of columns which do not have a validity code
            i2exp=i1+icolconc-ioffset-1
            if(i2.GE.i2exp) then
               call VGET320(char320,i1,i2,icolconc,ioffset,lvalid)
               if(.NOT.LVALID) rconc=-1.
            endif
         endif

c ---    Scale valid concentrations for units conversion
         if(rconc.GT.0.) rconc=t2_scale_ppb*rconc              

      endif

c --- Set records with small concentrations to zero
      if(rconc.GT.zero .AND. rconc.LT.cminppb) rconc=zero

      ieof=0
      return

999   ieof=1
      return
      end

c----------------------------------------------------------------------
      subroutine VGET320(c320,i1,i2,icol,ioffset,lvalid)
c----------------------------------------------------------------------
c
c --- PREPAQ    Version: 1.18     Level: 100108                VGET320
c ---           D. Strimaitis
c
c --- PURPOSE:  Interprets data validity character
c
c --- UPDATES:
c --- Version 1.15, Level: 081117 to Version 1.16, Level: 100108 (CEC)
c         - A character in non standard ASCII is used for "space".
c           in AMMNET files format. A Fix is to accept as VALID 
c           any FLAGS characters which are not unvalid.
c
c --- INPUTS:
c
c             C320 - character*320     - Input character string
c               i1 - integer           - First character of substring
c               i2 - integer           - Last character of substring
c             ICOL - integer           - Column for target data
c          IOFFSET - integer           - Offset to map data column to
c                                        position of character in
c                                        substring
c
c --- OUTPUT:
c           LVALID - logical           - Validity (T/F)
c
c --- VGET320 called by: (utility)
c --- VGET320 calls:      none
c----------------------------------------------------------------------
      parameter (mxvalid=12)

      character*1 vgood(mxvalid),vbad(mxvalid),c1
      character*320 c320
      logical lvalid

c --- Set validity flags
      data ngood/10/
      data vgood/' ','D','B','+','-','R','P','C','Q','x',2*''/
      data nbad/7/
      data vbad/'c','p','d','b','m','l','r',5*''/

      iqa=0

c --- Set character position in substring
      k=icol-ioffset
      k=i1+k-1

c --- Assign flag
      c1=c320(k:k)

c --- If position k exceeds end i2, then code should be a blank
c --- because trailing blanks may have been excluded in LEN_TRIM call
      if(k.GT.i2 .AND. c1.NE.' ') then
         write(*,*)'VGET: Invalid substring ranges'
         write(*,*)'      Starting position = ',i1
         write(*,*)'     Character position = ',k
         write(*,*)'        Ending position = ',i2
         write(*,*)'     Character(k) found = ',c1
         klast=LEN_TRIM(c320)
         write(*,*)c320(1:klast)
         stop
      endif

c --- Check Good Flags
      do i=1,ngood
         if(c1.EQ.vgood(i)) then
            iqa=i
            lvalid=.TRUE.
         endif
      enddo
      if(iqa.GT.0) return

c --- Check Bad Flags
      do i=1,nbad
         if(c1.EQ.vbad(i)) then
            iqa=i
            lvalid=.FALSE.
         endif
      enddo
      if(iqa.GT.0) return

c --- count warning:
	iw=iw+1
        if(iw.eq.50) then
        write(*,*)'Maximum number of warning messages reached'
        write(*,*)'No additional warnings will be written'
        elseif(iw.lt.50) then
c --- Flag assigned is not in the list
        write(*,*)'WARNING VGET: Accept Invalid flag found: ',c1
        endif
      lvalid=.TRUE.
c      stop

      return
      end

c----------------------------------------------------------------------
      subroutine readcf
c----------------------------------------------------------------------
c
c --- PREPAQ     Version: 1.18       Level: 091005               READCF
c                D. Strimaitis
c
c --- PURPOSE:  Read the file containing the file names of the
c               input and output files of the run, and associated
c               control variables, place into program variables,
c               and QA information.
c
c --- UPDATES:
c --- v1.13 (090310) to v1.14 (091005) DGS
c         - Add MVALID option for processing validity flags
c
c --- v1.12 (081205) to v1.13 (090310) DGS
c         - Add MPROC option for processing method
c
c --- v1.1 (081017) to v1.12 (081205) DGS
c         - Add concentration minimum
c         - Change default name of output data file (FILETAQ,taqdat)
c
c --- INPUTS:
c
c ---    Common block /QA/ variables:
c           VER, LEVEL
c
c        Parameters: ICTL, ILST, IOMESG, MXVAR
c
c --- OUTPUT:
c
c ---    Common block /FILNAM/ variables:
c           runlst,aqdat,taqdat,lcfiles
c ---    Common block /CONTROL/ variables:
c           itype,mproc,mvalid,cminppb,ihr1,ihr2,idelpd1,idelpd2,
c           idstn,xkm,ykm,iutmzn,rlat0,rlon0,rlat1,rlat2,feast,fnorth,
c           lgeo,lutm,llcc,lps,lem,llaza,lttm,
c           cspec, stn_nam,
c           clatdeg,clondeg,clat0,clon0,clat1,clat2,
c           pmap,datum,utmhem
c
c --- READCF called by:  SETUP
c --- READCF calls:      READIN, FILCASE, JULDAY, QAYR4, YR4, OPENS
c----------------------------------------------------------------------
c
c --- Include parameter statements and commons
      include 'params.paq'
      include 'params.cal'
c
c --- Include common blocks
      include 'control.paq'
      include 'filnam.paq'
      include 'qa.paq'
c
c --- Local variables
      character*4 ctemp(132,3), cstn_nam(40)
      character*4 clatlon(16,6), caspec(12)
      character*4 cabtz(8),cpmap(8),cdatum(8)
      character*12 cvdic(mxvar,3)
      integer ivleng(mxvar,3),ivtype(mxvar,3)
      logical lecho
      logical lerrcf

c --- Set control file error logical
      lerrcf=.FALSE.

c --- Set Dictionary
      data lecho/.false./
      data names/3/
      data cvdic/
     a  'FILEAQ','FILELST','FILETAQ','LCFILES', 56*' ',
     b  'TYPE','CSPEC','IHR1','IHR2','IDELPD1','IDELPD2',
     b  'CMIN','MPROC','MVALID', 51*' ',
     c  'IDSTN','STN_NAM','ABTZ','LATITUDE','LONGITUDE',
     c  'X_EASTING','Y_NORTHING','PMAP','FEAST','FNORTH',
     c  'IUTMZN','UTMHEM','RLAT0','RLON0','XLAT1','XLAT2',
     c  'DATUM', 43* ' '/

      data ivleng/
     a  3*132,1, 56*0,
     b  1,12,7*1, 51*0,
     c  1,40,8,2*16,2*1,8,4*1,4*16,8, 43*0/

c --- Variable types (ivtype) are coded as:
c          0 = null
c          1 = real
c          2 = integer
c          3 = logical
c          4 = character
      data ivtype/
     a  3*4,3, 56*0,
     b  2,4,4*2,1,2*2, 51*0,
     c  2,4*4,2*1,4,2*1,2,6*4, 43*0/

c ------------------
c --- File format 
c ------------------
c --- First line includes dataset types, version number
c --- and description
      read(ictl,'(2a16,a64)') inputset,inputver,inputmod
      
c ------------------
c --- Input Group 0
c ------------------

c --- Initialize the temporary arrays for filenames in Group 0
      do i=1,names
         do j=1,132
            ctemp(j,i)(1:1)=' '
         enddo
      enddo

c --- Read the group data
       call READIN(cvdic(1,1),ivleng(1,1),ivtype(1,1),ictl,iomesg,
     & lecho,
     1 ctemp(1,1),ctemp(1,2),ctemp(1,3),lcfiles,
     2 idum,idum,idum,idum,idum,idum,idum,idum,
     3 idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,
     4 idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,
     5 idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,
     6 idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum)

c --- Prepare any filenames included in the I/O file by erasing
c --- the default filename set above
      if(ctemp(1,1)(1:1).ne.' ')aqdat=' '
      if(ctemp(1,2)(1:1).ne.' ')runlst=' '
      if(ctemp(1,3)(1:1).ne.' ')taqdat=' '
c --- Transfer the char*4 data into the char*132 variables
      do j=1,132
         if(ctemp(j,1)(1:1).ne.' ')aqdat(j:j)=ctemp(j,1)(1:1)
         if(ctemp(j,2)(1:1).ne.' ')runlst(j:j)=ctemp(j,2)(1:1)
         if(ctemp(j,3)(1:1).ne.' ')taqdat(j:j)=ctemp(j,3)(1:1)
      enddo

c --- Convert the file names to the proper case
      call FILCASE(lcfiles,aqdat)
      call FILCASE(lcfiles,runlst)
      call FILCASE(lcfiles,taqdat)

c --- Open listfile
      open(ilst,file=runlst,status='unknown')

c --- Write banner to list file
      write(ilst,5) ver,level
5     format(///,26x,'PREPAQ OUTPUT SUMMARY',/,19x,'VERSION:  ',A8,
     1       ' LEVEL:  ',A8///)

c -----------------
c --- Input Group 1
c -----------------

c --- Initialize the species array
      do j=1,12
         caspec(j)(1:1)=' '
         cspec(j:j)=' '
      enddo

      call readin(cvdic(1,2),ivleng(1,2),ivtype(1,2),ictl,ilst,lecho,
     1 ITYPE,CASPEC,IHR1,IHR2,IDELPD1,IDELPD2,CMINPPB,MPROC,MVALID,
     2 idum,idum,idum,
     3 idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,
     4 idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,
     5 idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,
     6 idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum)

c --- Transfer the char*4 data into the char*12 variable (species)
      do j=1,12
         if(caspec(j)(1:1).ne.' ')cspec(j:j)=caspec(j)(1:1)
      enddo

c -----------------
c --- Input Group 2
c -----------------

c --- Initialize the temporary array for the station name 
      do j=1,40
         cstn_nam(j)(1:1)=' '
         stn_nam(1:1)=' '
      enddo

c --- Initialize the temporary array for the time zone
c --- map projection
      do j=1,8
         cabtz(j)(1:1)=' '
         abtz(1:1)=' '
      enddo

c --- Initialize the temporary arrays for the character lat/lon fields
      do i=1,6
         do j=1,16
            clatlon(j,i)(1:1)=' '
         enddo
      enddo
      do j=1,16
         clatdeg(j:j)=' '
         clondeg(j:j)=' '
         clat0(j:j)=' '
         clon0(j:j)=' '
         clat1(j:j)=' '
         clat2(j:j)=' '
      enddo

c --- Initialize the temporary array for the Datum-Region name and 
c --- map projection
      do j=1,8
         cpmap(j)(1:1)=' '
         cdatum(j)(1:1)=' '
      enddo

      call READIN(cvdic(1,3),ivleng(1,3),ivtype(1,3),ictl,ilst,lecho,
     1 IDSTN,CSTN_NAM,CABTZ,CLATLON(1,1),CLATLON(1,2),XKM,YKM,
     2 CPMAP,FEAST,FNORTH,IUTMZN,UTMHEM,CLATLON(1,3),CLATLON(1,4),
     3 CLATLON(1,5),CLATLON(1,6),CDATUM,
     4 idum,idum,idum,idum,idum,idum,idum,
     5 idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,
     6 idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,
     7 idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum,idum)


c --- Transfer the char*4 data into the char*40 variables
      do j=1,40
         stn_nam(j:j)=cstn_nam(j)(1:1)
      enddo

c --- Transfer the char*4 data into the char*16 variables
      do j=1,16
         if(clatlon(j,1)(1:1).ne.' ')clatdeg(j:j)=clatlon(j,1)(1:1)
         if(clatlon(j,2)(1:1).ne.' ')clondeg(j:j)=clatlon(j,2)(1:1)
         if(clatlon(j,3)(1:1).ne.' ')clat0(j:j)=clatlon(j,3)(1:1)
         if(clatlon(j,4)(1:1).ne.' ')clon0(j:j)=clatlon(j,4)(1:1)
         if(clatlon(j,5)(1:1).ne.' ')clat1(j:j)=clatlon(j,5)(1:1)
         if(clatlon(j,6)(1:1).ne.' ')clat2(j:j)=clatlon(j,6)(1:1)
      enddo

c --- Transfer the char*4 data into the char*8 variables
      do j=1,8
         abtz(j:j)=cabtz(j)(1:1)
      enddo
      if(cpmap(1)(1:1).ne.' ') then
         do j=1,8
            pmap(j:j)=cpmap(j)(1:1)
         enddo
      endif
      if(cdatum(1)(1:1).ne.' ') then
         do j=1,8
            datum(j:j)=cdatum(j)(1:1)
         enddo
      endif

c --- Pad the char*4 UTM Hemisphere
      utmhem(2:4)='   '

c     QA
c -------------------------

c --- File Type
      if(itype.LT.1 .OR. itype.GT.2) then
         write(ilst,*)
         write(ilst,*) '*****'
         write(ilst,*) 'ERROR in control file: bad File Type = ',itype
         write(ilst,*) '*****'
         write(ilst,*)
         LERRCF=.TRUE.
      endif
c --- Processing method
      if(mproc.LT.0 .OR. mproc.GT.1) then
         write(ilst,*)
         write(ilst,*) '*****'
         write(ilst,*) 'ERROR in control file: bad Processing Method'
         write(ilst,*) '     Expected MPROC = 0,1'
         write(ilst,*) '     Found    MPROC = ',mproc
         write(ilst,*) '*****'
         write(ilst,*)
         LERRCF=.TRUE.
      endif
c --- Validity Processing method
      if(mvalid.LT.0 .OR. mvalid.GT.1) then
         write(ilst,*)
         write(ilst,*) '*****'
         write(ilst,*) 'ERROR in control file: bad Validity Method'
         write(ilst,*) '     Expected MVALID = 0,1'
         write(ilst,*) '     Found    MVALID = ',mvalid
         write(ilst,*) '*****'
         write(ilst,*)
         LERRCF=.TRUE.
      endif

c --- Drop trajectory time-range when processing all concentrations
      if(mproc.EQ.1) then
         idelpd1=0
         idelpd2=0
      endif

c --- Projection
      if(pmap.EQ.'LL') then
         lgeo=.TRUE.
      elseif(pmap.EQ.'UTM') then
         lutm=.TRUE.
      elseif(pmap.EQ.'LCC') then
         llcc=.TRUE.
      elseif(pmap.EQ.'PS') then
         lps=.TRUE.
      elseif(pmap.EQ.'EM') then
         lem=.TRUE.
      elseif(pmap.EQ.'LAZA') then
         llaza=.TRUE.
      elseif(pmap.EQ.'TTM') then
         lttm=.TRUE.
      else
         write(ilst,*)
         write(ilst,*) '*****'
         write(ilst,*) 'ERROR in control file: bad PMAP = ',pmap
         write(ilst,*) '*****'
         write(ilst,*)
         LERRCF=.TRUE.
      endif

c ------------------------
c --- Open remaining files
c ------------------------
      open(icsv,file=aqdat,status='old')
      open(iout,file=taqdat,status='unknown')

c ----------------------------
c --- Echo inputs to list file
c ----------------------------

c --- Echo filenames
      write(ilst,*)
      write(ilst,*) 'Control file name:     ',runinp
      write(ilst,*) 'AQ Data file name:     ',aqdat
      write(ilst,*) 'Output list file name: ',runlst
      write(ilst,*) 'Output file name:      ',taqdat

      write(ilst,*)
      write(ilst,*) '  Column Names ---'
      write(ilst,*) '         Timestamp: ', COL_DTM
      write(ilst,*) '  Averaging period: ', COL_AVG
      write(ilst,*) '        Station ID: ', COL_STN

      write(ilst,*)
      write(ilst,*) 'Group 1 Inputs'
      write(ilst,*) '   Input File Type: ', itype
      write(ilst,*) '           Species: ', cspec
      write(ilst,*) ' Processing Method: ', mproc
      write(ilst,*) 'Validity Code Used: ', mvalid
      write(ilst,*) 'Minimum Conc (ppb): ', cminppb
      write(ilst,*) 'Hour Search Window: ', ihr1,ihr2
      write(ilst,*) '      Period range: ', idelpd1,idelpd2

      write(ilst,*)
      write(ilst,*) 'Group 2 Inputs'
      write(ilst,*) '        Station ID: ', idstn
      write(ilst,*) '      Station Name: ', stn_nam
      write(ilst,*) '         Time Zone: ', abtz
      write(ilst,*) '    Map Projection: ', pmap
      write(ilst,*) '             Datum: ', datum

      if(LGEO) then
         write(ilst,*)
         write(ilst,*) 'Lat/Lon prescription of location'
         write(ilst,*) '          Latitude: ', clatdeg
         write(ilst,*) '         Longitude: ', clondeg
      else
         write(ilst,*)
         write(ilst,*) 'X/Y prescription of location'
         write(ilst,*) '             x(km): ', xkm
         write(ilst,*) '             y(km): ', ykm
      endif

      if(LUTM) then
         write(ilst,*)
         write(ilst,*) '          UTM Zone: ', iutmzn
         write(ilst,*) '        Hemisphere: ', utmhem
      elseif(.NOT.LGEO) then
         write(ilst,*)
         write(ilst,*) '     False Easting: ', feast
         write(ilst,*) '    False Northing: ', fnorth
         write(ilst,*) '         Ref. Lat.: ', clat0
         write(ilst,*) '         Ref. Lon.: ', clon0
         write(ilst,*) '        Match Lat.: ', clat1
         write(ilst,*) '        Match Lat.: ', clat2
      endif

c --- STOP now if error exists in the control file
      if(LERRCF) then
         write(*,*)'ERRORS are found in the CONTROL file'
         write(*,*)'Review messages written to the LIST file'
         stop
      endif

      return
      end

c----------------------------------------------------------------------
      subroutine comp
c----------------------------------------------------------------------
c
c --- PREPAQ    Version: 1.18     Level: 091016                   COMP
c ---           D. Strimaitis
c
c --- PURPOSE:  Main computational component of program
c
c --- UPDATES:
c --- v1.13 (090310) to v1.15 (091016) DGS
c         - Distinguish between missing and small concentrations and
c           report cases with no data that meet selection criteria
c --- v1.11 (081205) to v1.13 (090310) DGS
c         - Add MPROC option for processing method (Peak or All)
c         - Refine/correct processing window logic
c
c --- INPUTS:
c
c ---    Common block /CONTROL/ variables:
c           itype,mproc,ihr1,ihr2,idelpd1,idelpd2,idstn,xkm,ykm,
c           iutmzn,rlat0,rlon0,rlat1,rlat2,feast,fnorth,
c           lgeo,lutm,llcc,lps,lem,llaza,lttm,
c           cspec, stn_nam,
c           clatdeg,clondeg,clat0,clon0,clat1,clat2,
c           pmap,datum,utmhem
c ---    Common block /QA/ variables:
c           ver,level
c ---    Common block /AQINPUT/ variables:
c           iavg,istn,iyr,jul,ihr,isec,rconc
c
c --- OUTPUT:
c             none
c
c --- COMP called by: (main)
c
c --- COMP calls:      READPD, INCRS, GRDAY
c----------------------------------------------------------------------     
c
c --- Include parameter statements and commons
      include 'params.paq'
c
c --- Include common blocks
      include 'control.paq'
      include 'qa.paq'
      include 'aqinput.paq'

      character*80 aline80, blnk80
      character*40 appb
      character*1 q
      real zero
      logical loutput
c --- Flags for reporting nature of problems found
      logical lnotime, lmissing, lessthresh

      data lnotime/.TRUE./,lmissing/.TRUE./,lessthresh/.TRUE./
      data zero/0.0/

c --- Report progress
      write(iomesg,*)'COMPUTATIONAL PHASE'

c --- Create a start-time for range checks
      ihr0=ihr1-1
      ihr0=MAX(ihr0,0)
      isec0=0
      istart=ihr0*10000+isec0

c --- Create an end-time for range checks
      isec2=0
      istop=ihr2*10000+isec2

c --- Trim species label
      kspec=LEN_TRIM(cspec)

c --- Set data record label
      q=''''
      if(mproc.EQ.0) then
         appb=q//'Peak '//cspec(1:kspec)//'       ppb'//q
      elseif(mproc.EQ.1) then
         appb=q//cspec(1:kspec)//'       ppb'//q
      endif
      kappb=LEN_TRIM(appb)
      ic1=kappb-9
      ic2=kappb-5

      do k=1,80
         blnk80(k:k)=' '
      enddo

c --- Header records
      write(iout,'(2a16,a40)') 'TRAJLOC.DAT     ','1.0             ',
     &           'Time zone, Coord information            '
      write(iout,'(i1)') 2
      write(iout,'(a)') 'Created by PREPAQ Version '//ver//level
      write(iout,'(a70)') 'Time-ranges and labels associated with '//
     &                    'tractories from ONE location   '

      write(iout,'(a8)') pmap
      if(LUTM) then
         write(iout,'(i4,a4)') iutmzn,utmhem
      elseif(LLCC) then
         write(iout,'(4a16)') clat0,clon0,clat1,clat2
      elseif(LPS) then
         write(iout,'(3a16)') clat0,clon0,clat1
      elseif(LEM.or.LLAZA.or.LTTM) then
         write(iout,'(2a16)') clat0,clon0
      endif
      if(LLCC.or.LLAZA.or.LTTM) then
         write(iout,*) feast,fnorth
      endif
      write(iout,'(a8,a12)') datum,daten

      if(LGEO) then
         write(iout,'(a4)') 'DEG '
      else
         write(iout,'(a4)') 'KM  '
      endif

      write(iout,'(a8)') abtz
      write(iout,'(a40)') stn_nam

      if(LGEO) then
         write(iout,'(2a16)') clatdeg,clondeg
      else
         write(iout,'(2f13.4)') xkm,ykm
      endif

c --- Initialize
      cmax=-0.1

c --- Process data records
20    call READPD(ieof)
      if(ieof.GT.0) then
         write(ilst,*)
         write(ilst,*)'Completed all records in AQ data file'
         if(mproc.EQ.0) then
            goto 30
         else
            goto 40
         endif
      endif

c --- Check data record for target station within time window
      if(istn.NE.idstn) goto 20
c --- Set the end-time of the measurement period
      itime=ihr*10000+isec
      if(itime.EQ.0) then
c ---    Treat special case of period that ends at midnight
         itime=240000
      endif
c --- Skip periods that end at the start time or earlier
      if(itime.LE.istart) goto 20

c --- QA reporting
c --- Time within window?
      if(itime.LE.istop) then
         lnotime=.FALSE.
c ---    Value not missing?
         if(rconc.GE.zero) lmissing=.FALSE.
c ---    Value meets threshold?
         if(rconc.GT.zero) lessthresh=.FALSE.
      endif

c --- Place information into output variables
      if(mproc.EQ.0) then
c ---    Update information for peak concentration if
c ---    period is before the end of the window
         if(itime.LE.istop .AND. rconc.GT.cmax) then
            mxyr=iyr
            mxjul=jul
            mxhr=ihr
            mxsec=isec
            cmax=rconc
         endif
      elseif(mproc.EQ.1) then
c ---    Update information for ALL concentrations if
c ---    period is before the end of the window
         if(itime.LE.istop .AND. rconc.GT.0.) then
            mxyr=iyr
            mxjul=jul
            mxhr=ihr
            mxsec=isec
            cmax=rconc
         endif
      endif

30    loutput=.FALSE.
      if(mproc.EQ.0) then
         if((ieof.GT.0 .OR. itime.GE.istop) .AND. cmax.GT.0.0)
     &                    loutput=.TRUE.
      elseif(mproc.EQ.1) then
         if(cmax.GT.0.0) loutput=.TRUE.
      endif

      if(loutput) then
c ---    Prepare output record
c ---    Peak concentration string
         mxppb=NINT(cmax)
         write(appb(ic1:ic2),'(i5)') mxppb
c ---    Begin time
         ibyr=mxyr
         ibjul=mxjul
         ibhr=mxhr
         ibsec=mxsec
         nsecinc=-idelpd1*iavg
         call INCRS(ilst,ibyr,ibjul,ibhr,ibsec,nsecinc)
         call GRDAY(ilst,ibyr,ibjul,ibmo,ibday)
c ---    End time
         ieyr=mxyr
         iejul=mxjul
         iehr=mxhr
         iesec=mxsec
         nsecinc=idelpd2*iavg
         call INCRS(ilst,ieyr,iejul,iehr,iesec,nsecinc)
         call GRDAY(ilst,ieyr,iejul,iemo,ieday)

c ---    Write record
         write(iout,'(2(i4.4,3i4.2,i6.4,4x),a)')
     &                    ibyr,ibmo,ibday,ibhr,ibsec,
     &                    ieyr,iemo,ieday,iehr,iesec,appb(1:kappb)
c ---    Reset max concentration
         cmax=-1.0
      endif

      if(ieof.EQ.0) goto 20

c --- Section to report QA messages
c --- Write completion notice to list file
40    write(ilst,*)
      write(ilst,*)
      if(lnotime) then
         write(ilst,*)'ERROR: No data records are in selected period'
      elseif(lmissing) then
         write(ilst,*)'ERROR: All selected concentrations are missing'
      elseif(lessthresh) then
         write(ilst,*)'ERROR: Selected threshold is not exceeded'
      else
         write(ilst,*)'RUN COMPLETED SUCCESSFULLY'
      endif

      return
      end

c----------------------------------------------------------------------
      subroutine fin(itest)
c----------------------------------------------------------------------
c
c --- PREPAQ     Version: 1.18        Level: 091016                 FIN
c ---            J. Scire, D. Strimaitis

c
c --- PURPOSE:  Run termination routine -- compute runtime,
c               write last day processed
c
c --- UPDATES:
c --- v1.1 (081017) to v1.15 (091016) DGS
c         - Replace file unit number io6 with iomesg and report time to
c           list file
c
c --- INPUTS:
c          ITEST - integer - Flag indicating if execution is to
c                            include COMPUTATIONAL phase
c                            (ITEST = 1 to STOP program and skip
c                                       the COMPUTATIONAL phase
c                             ITEST = 2 to CONTINUE execution to
c                                       include computations)
c       Common block /DATEHR/ ## Not used here
c          jyr, jjul, jhr
c       Common block /QA/
c          rdate, rtime, rcpu
c       Parameters: ILST, IOMESG
c
c --- OUTPUT:  none
c
c --- FIN called by:  MAIN
c --- FIN calls:      DATETM, JULDAY, DELTT, YR4C
c----------------------------------------------------------------------
c
c --- include parameters
      include 'params.paq'
c
      character*8 rtime2
      character*10 rdate2
c
c##      include 'datehr.paq'
      include 'qa.paq'
c
      write(iomesg,*)'TERMINATION PHASE'
c
cc --- Write last day/hour processed
c      if(ITEST.eq.2)then
cc ---    Compute month & day from Julian day
c         call GRDAY(io6,jyr,jjul,jmo,jday)
c         write(io6,5)jyr,jmo,jday,jjul,jhr
c5        format(//2x,'LAST DAY/HOUR PROCESSED:'/5x,'Year: ',i4,2x,
c     1   'Month: ',i2,3x,'Day: ',i2,3x,'Julian day: ',i3,3x,'Hour: ',
c     2   i2)
c      else
cc
cc ---    TEST mode -- COMPUTATIONAL phase skipped
c         write(io6,12)
c12       format(/1x,13('----------')//1x,
c     1   'Completion of test mode run -- run terminating ',
c     2   'normally'//1x,13('----------'))
c      endif

c
c --- get system date & time at end of run
      call datetm(rdate2,rtime2,rcpu)
c
c --- compute runtime
      read(rtime(1:2),10)ihr1
      read(rtime(4:5),10)imin1
      read(rtime(7:8),10)isec1
10    format(i2)
      t1=ihr1*3600.+imin1*60.+isec1
c
      read(rtime2(1:2),10)ihr2
      read(rtime2(4:5),10)imin2
      read(rtime2(7:8),10)isec2
      t2=ihr2*3600.+imin2*60.+isec2
c
      if(rdate.eq.rdate2)then
         delt=t2-t1
      else
         read(rdate(1:2),10)imo1
         read(rdate(4:5),10)iday1
         read(rdate(7:10),'(i4)')iyr1
         call julday(iomesg,iyr1,imo1,iday1,ijul1)

         read(rdate2(1:2),10)imo2
         read(rdate2(4:5),10)iday2
         read(rdate2(7:10),'(i4)')iyr2
         call julday(iomesg,iyr2,imo2,iday2,ijul2)

c ---    compute no. hours from beg. of first hour of run to
c ---    ending hour of ending day of the run
         call deltt(iyr1,ijul1,ihr1,iyr2,ijul2,ihr2,idelhr)

c ---    adjust for minutes and seconds
         delt=idelhr*3600.-imin1*60.-isec1+imin2*60.+isec2
      endif

c --- On the PC, the runtime and CPU time are the same
c --- (DATETM provides RCPU = 0.0 on the PC)
      if(rcpu.EQ.0.0)rcpu=delt

c --- Report current date to screen and list file
      write(iomesg,1402)rtime2,rdate2,delt,rcpu
      write(ilst,1402)rtime2,rdate2,delt,rcpu
1402  format(//2x,'End of run -- Clock time: ',a8/
     1         2x,'                    Date: ',a10//
     2         2x,'      Elapsed clock time: ',f10.1,' (seconds)'//
     3         2x,'                CPU time: ',f10.1,' (seconds)')
c
      return
      end

c----------------------------------------------------------------------
      subroutine case320(mode,c320,nlim)
c----------------------------------------------------------------------
c
c --- PREPAQ    Version: 1.18     Level: 081017                CASE320
c ---           D. Strimaitis
c
c --- PURPOSE:  Convert all letters within a character string to 
c               either upper or lower case
c
c --- INPUTS:
c
c             MODE - character*5       - Output directive for either 
c                                        upper or lower case
c             C320 - character*320     - Input character string
c             NLIM - integer           - Length of string (characters)
c
c --- OUTPUT:
c
c             C320 - character*320     - Output character string with
c                                        uniform case
c
c --- CASE320 called by: (utility)
c --- CASE320 calls:      IACHAR, CHAR
c----------------------------------------------------------------------

      character*320 c320
      character*5 mode

      if(nlim.GT.320) then
         write(*,*)'ERROR in Subr. CASE320 - string is too long'
         write(*,*)'Maximum length  : 320'
         write(*,*)'Requested length: ',nlim
         stop 'Program HALTED'
      endif

      if(mode.EQ.'UPPER' .OR. mode.EQ.'upper') then
         do k=1,nlim
            ia=IACHAR(c320(k:k))
            if(ia.GE.97 .AND. ia.LE.122) then
               ia=ia-32
               c320(k:k)=CHAR(ia)
            endif
         enddo

      elseif(mode.EQ.'LOWER' .OR. mode.EQ.'lower') then
         do k=1,nlim
            ia=IACHAR(c320(k:k))
            if(ia.GE.65 .AND. ia.LE.90) then
               ia=ia+32
               c320(k:k)=CHAR(ia)
            endif
         enddo
      else
         write(*,*)'ERROR in Subr. CASE320 - invalid output mode'
         write(*,*)'Expected UPPER or LOWER'
         write(*,*)'Found: ',mode
         stop 'Program HALTED'
      endif

      return
      end

c----------------------------------------------------------------------
      subroutine noblnk320(c320)
c----------------------------------------------------------------------
c
c --- PREPAQ    Version: 1.18     Level: 081017              NOBLNK320
c ---           D. Strimaitis
c
c --- PURPOSE:  Remove all blanks from string (left-justify)
c
c --- INPUTS:
c
c             C320 - character*320     - Input character string
c
c --- OUTPUT:
c
c             C320 - character*320     - Output character string with
c
c --- NOBLNK320 called by: (utility)
c --- NOBLNK320 calls:      ADJUSTL, LEN_TRIM
c----------------------------------------------------------------------

      character*320 c320,d320

c --- Blank out work string
      do i=1,320
         d320(i:i)=' '
      enddo

c --- Remove leading blanks
      c320=ADJUSTL(c320)
      last=LEN_TRIM(c320)

c --- Transfer non-blanks into work string
      n=0
      do i=1,last
         if(c320(i:i).NE.' ') then
            n=n+1
            d320(n:n)=c320(i:i)
         endif
      enddo

c --- Pass work string
      c320=d320

      return
      end

c----------------------------------------------------------------------
      subroutine COLUMN320(c320,nlim,cd,ncol,nc,ic)
c----------------------------------------------------------------------
c
c --- PREPAQ    Version: 1.18     Level: 081117              COLUMN320
c ---           D. Strimaitis
c
c --- PURPOSE:  Identifies location of all column delimiters in a
c               string and checks for expected number
c
c --- INPUTS:
c
c             C320 - character*320     - Input character string
c             NLIM - integer           - Length of string (characters)
c               CD - character*1       - Column delimiter character
c             NCOL - integer           - Number of columns expected
c                                        (Enter 99 if unknown)
c
c --- OUTPUT:
c
c             NCOL - integer           - Number of columns processed
c               NC - integer           - Number of delimiters
c       IC(0:NCOL) - integer           - Location of delimiters
c
c --- COLUMN320 called by: (utility)
c --- COLUMN320 calls:      none
c----------------------------------------------------------------------

      character*1 cd
      character*320 c320
      integer ic(0:ncol)
      logical lncol

      if(nlim.GT.320) then
         write(*,*)'ERROR in Subr. COLUMN320 - string is too long'
         write(*,*)'Maximum length  : 320'
         write(*,*)'Requested length: ',nlim
         stop 'Program HALTED'
      endif

c --- Set logical for processing a specific number of columns
      if(ncol.GT.99) then
         write(*,*)'ERROR in Subr. COLUMN320 - too many columns'
         write(*,*)'Maximum number  : 99'
         write(*,*)'Number requested: ',ncol
         stop 'Program HALTED'
      endif
      lncol=.FALSE.
      if(ncol.LT.99) lncol=.TRUE.

      nc=0
      ic(0)=0
      do k=1,nlim
         if(c320(k:k).EQ.cd) then
            if(LNCOL) then
c ---          Actual number known
               if(nc.LT.ncol) then
                  nc=nc+1
                  ic(nc)=k
               endif
            else
c ---          Actual number unknown
               nc=nc+1
               ic(nc)=k
               ncol=nc
            endif
         endif
      enddo
c --- Last delimiter in line is usually missing
      if(LNCOL) then
         if(nc.LT.ncol) then
            nc=nc+1
            ic(nc)=nlim+1
         endif
      else
         nc=nc+1
         ic(nc)=nlim+1
         ncol=nc
      endif
c --- Check number of columns
      if(nc.LT.ncol) then
         write(*,*)'ERROR in Subr. COLUMN320 - '
         write(*,*)'Too few columns in the line'
         write(*,*)'Expected number of columns: ',ncol
         write(*,*)'                     Found: ',nc
         write(*,'(a)')c320(1:nlim)
         stop 'Program HALTED'
      endif

      return
      end

c----------------------------------------------------------------------
      subroutine IRANGE(n,ic,nc,i1,i2)
c----------------------------------------------------------------------
c
c --- PREPAQ    Version: 1.18     Level: 081017                 IRANGE
c ---           D. Strimaitis
c
c --- PURPOSE:  Extracts starting and ending position of field for a
c               particular column in a CSV line
c
c --- INPUTS:
c
c                N - integer           - Column number for field
c         IC(0:NC) - integer           - Location of commas
c               NC - integer           - Number of commas
c
c --- OUTPUT:
c
c               I1 - integer           - Starting position of field
c               I2 - integer           - Ending position of field
c
c --- IRANGE called by: (utility)
c --- IRANGE calls:      none
c----------------------------------------------------------------------

      integer ic(0:nc)

      i1=ic(n-1)+1
      i2=ic(n)-1

      return
      end

c----------------------------------------------------------------------
      subroutine rget320(c320,i1,i2,x)
c----------------------------------------------------------------------
c
c --- PREPAQ    Version: 1.18     Level: 100128                RGET320
c ---           D. Strimaitis
c
c --- PURPOSE:  Convert character string to a real.  This sub is needed
c               when there is no decimal point.  The LF95 compiler
c               does not do the internal read conversion without the '.'
c               Also, determine when exponential notation is present.
c
c --- UPDATES:
c --- Version 1.16 (level 081017) to Version 1.17 (level 100128) (CEC)
c         - Add a check when any possible concentration is empty
c           or have already missing parameters such as -6999. or 9999. or -9999.
c
c --- INPUTS:
c
c             C320 - character*320     - Input character string
c               i1 - integer           - First character of substring
c               i2 - integer           - Last character of substring
c
c --- OUTPUT:
c
c                X - real              - Output real
c
c --- RGET320 called by: (utility)
c --- RGET320 calls:      none
c----------------------------------------------------------------------
      character*320 c320
      character*20 c20, blnk20
      data blnk20/'                    '/

c --- If i1 is larger than i2, it means the column is empty
c --- Value is then put as missing = 9999.
      if(i1.GT.i2) then
      x=9999.
      return
      endif

      c20=blnk20
      c20=c320(i1:i2)

c --- Left justify field
      c20=ADJUSTL(c20)

c --- Test for exponential notation
      ie=MAX(INDEX(c20,'e'),INDEX(c20,'E'))

      if(ie.EQ.0) then
c ---    Simple real number
         ipt=INDEX(c20,'.')
         if(ipt.EQ.0) then
c ---       Decimal point NOT found
            ipt=LEN_TRIM(c20)+1
            c20(ipt:ipt)='.'
         endif
c ---    Decimal point in place: internal read OK
         read(c20,'(f)') x

      else
c ---    Exponential notation
         ipt=INDEX(c20,'.')
         if(ipt.EQ.0) then
c ---       Decimal point NOT found
            do k=20,ie+1,-1
               km1=k-1
               c20(k:k)=c20(km1:km1)
            enddo
            c20(ie:ie)='.'
         endif
c ---    Decimal point in place: internal read OK
         read(c20,'(e)') x

      endif

c --- check on possible other missing notations
      if (x.gt.9998..or.x.lt.-6998) then
      x=9999.
      endif

      return
      end

c----------------------------------------------------------------------
      subroutine omit320(char1,char320)
c----------------------------------------------------------------------
c
c --- PREPAQ    Version: 1.18     Level: 081017                OMIT320
c ---           D. Strimaitis
c
c --- PURPOSE:  Removes all instances of a character from a string
c
c --- INPUTS:
c            CHAR1 - char*1       - Character removed
c          CHAR320 - char*320     - Current content of variable
c                                   assignment string
c
c --- OUTPUT:
c          CHAR320 - char*320     - Updated content of variable
c                                   assignment string
c
c --- OMIT320 called by: (main)
c --- OMIT320 calls:     none
c----------------------------------------------------------------------
      character*1 char1
      character*320 char320,work320

      do i=1,320
         work320(i:i)=' '
      enddo

      kk=0
      klast=LEN_TRIM(char320)
      do k=1,klast
         if(char320(k:k).NE.char1) then
            kk=kk+1
            work320(kk:kk)=char320(k:k)
         endif
      enddo
      char320=work320

      return
      end

c ---------------------------------------------------------------------
      subroutine rd_datim(i1,i2,datm,imo,iday,iyr,ihour,irmn)
c ---------------------------------------------------------------------
c
c --- PREPAQ    Version: 1.18     Level: 091016               RD_DATIM
c --- C. Escoffier           
c
c --- PURPOSE:  Read DATE/TIME from a worksheet string variable
c
c----------------------------------------------------------------------
c
        dimension idbsl(2)
        character*16 datm
c
         ibsl=0
	 do ich=i1,i2
            if(datm(ich:ich).eq.'/') then
            ibsl=ibsl+1
            idbsl(ibsl)=ich
            endif
            if(datm(ich:ich).eq.':') then
            idclu=ich
            endif
         enddo
         if(ibsl.gt.2) then
         write(*,*)'ERROR - date in wrong format'
         write(*,*)'Format expected: mm/dd/yyyy hh:mm'
         stop
         endif
        
         if(idbsl(1).eq.2) then
         read(datm(1:1),'(i1)')imo
         elseif(idbsl(1).eq.3) then
         read(datm(1:2),'(i2)')imo
         else
	 write(*,*)'error - date not in correct format'
         write(*,*)'format expected: mm/dd/yyyy hh:mm'
         stop
         endif

         if(idbsl(2)-idbsl(1).eq.2) then
         lbsl=idbsl(1)+1
         read(datm(lbsl:lbsl),'(i1)')iday
         elseif(idbsl(2)-idbsl(1).eq.3) then
         lbsl=idbsl(1)+1
         lbsl1=idbsl(1)+2
         read(datm(lbsl:lbsl1),'(i2)')iday
         else
         write(*,*)'error - date not in correct format'
         write(*,*)'format expected: mm/dd/yyyy hh:mm'
         stop
         endif

         lbsl=idbsl(2)+1
         lbsl1=idbsl(2)+4
         read(datm(lbsl:lbsl1),'(i4)')iyr

         if(idclu-idbsl(2).eq.7) then
         lbsl=idclu-1
         read(datm(lbsl:lbsl),'(i1)')ihour
         elseif(idclu-idbsl(2).eq.8) then
         lbsl=idclu-2
         lbsl1=idclu-1
         read(datm(lbsl:lbsl1),'(i2)')ihour
         else
         write(*,*)'error - date not in correct format'
         write(*,*)'format expected: mm/dd/yyyy hh:mm'
         stop
         endif
         
         lbsl=idclu+1
         lbsl1=idclu+2   
         read(datm(lbsl:lbsl1),'(i2)') irmn

	 return
         end

